#!/usr/bin/env python3

# Open the entry point when program is executed using 'md2anki [...]' (check setup.py for more information)

# Local modules
from md2anki.entry_point import entry_point


def _main():
    entry_point()
